;; Generated package description from kiwix.el  -*- no-byte-compile: t -*-
(define-package "kiwix" "1.1.0.0.20210724.123701" "Searching offline Wikipedia through Kiwix." '((emacs "24.4") (request "0.3.0") (elquery "0.1.0")) :keywords '("kiwix" "wikipedia") :authors '(("stardiviner" . "numbchild@gmail.com")) :maintainer '("stardiviner" . "numbchild@gmail.com") :url "https://github.com/stardiviner/kiwix.el")
